
#define currentGPU 0
#define ELEMENTS 5

// Constantes de los pesos del stencil
#define WEIGHT_CENTER   (float)  (0.5   * (1.0/1.0) )
#define WEIGHT_ADJACENT (float) (0.295 * (1.0/4.0) )
#define WEIGHT_CORNER   (float) (0.195 * (1.0/4.0) )
#define MAX_RESID       (float) 0.001
#define EXT_RESID       (float) MAX_RESID + 1.0


#ifndef _KERNELS_
#define _KERNELS_

///////////////////////////////////////////////////////////////////////////////////////////
//! Kernel que realiza la actualizacion de una matriz de entrada y lo deja en una matriz de salida
//! @param matrix_old      Matriz de los valores de entrada
//! @param matrix_new      Matrix de los valores de salida
//! @param rows            Numero de filas de las matrices
//! @param cols            Numero de columnas de las matrices
///////////////////////////////////////////////////////////////////////////////////////////
__global__ void gpuFunc_Update(m_type * matrix_old, m_type * matrix_new, int rows, int cols, m_type * resid){
        //************************************************************
        // 1. Macros para acceder a las matrices
        #define m_old(i,j) (matrix_old[(i)+(j)*(cols)])
        #define m_new(i,j) (matrix_new[(i)+(j)*(cols)])
        //************************************************************

	int IDX_Thread = threadIdx.x; 	//Identificación del hilo de la dimensión X
	int IDY_Thread = threadIdx.y; 	//Identificación del hilo de la dimensión Y
	int IDX_Block = blockIdx.x; 	//Identificación del hilo de la dimensión X
	int IDY_Block = blockIdx.y; 	//Identificación del hilo de la dimensión Y
	int shapeBlock_X = blockDim.x;	//Número de hilos por bloque en la dimensión X
	int shapeBlock_Y = blockDim.y;	//Número de hilos por bloque en la dimensión Y

	int tidX= IDX_Block * shapeBlock_X + IDX_Thread;		
	int tidY= IDY_Block * shapeBlock_Y + IDY_Thread;		

	int tidXMin = (tidX-1+cols)%cols;
	int tidYMin = (tidY-1+cols)%cols;
	int tidXMax = (tidX+1)%cols;
	int tidYMax = (tidY+1)%cols;

	m_new(tidX,tidY) = ( m_old (tidX,tidY) * WEIGHT_CENTER) +

                                                ((m_old (tidXMin,tidY) +
                                                        m_old (tidXMax,tidY) +
                                                        m_old (tidX,tidYMin) +
                                                        m_old (tidX,tidYMax) ) * WEIGHT_ADJACENT) +

                                                ((m_old (tidXMin,tidYMin) +
                                                        m_old (tidXMin,tidYMax) +
                                                        m_old (tidXMax,tidYMin) +
                                                        m_old (tidXMax,tidYMax) ) * WEIGHT_CORNER);

	//res(tidX,tidY)=fabs(m_old(tidX,tidY) - m_new(tidX,tidY));
	// Basta con que uno residuo sea mayor que el máximo permitido para repetir el bucle
	// Y no es necesario que el valor de la bandera "resid" tenga un valor concreto
	// basta con ser mayor que MAX_RESID
	if ((fabs(m_old(tidX,tidY) - m_new(tidX,tidY)))>MAX_RESID) resid[0]=EXT_RESID;

        #undef m_new
        #undef m_old	
}
        //************************************************************

///////////////////////////////////////////////////////////////////////////////////////////
//! Kernel que realiza la reducción de un array de entrada y lo deja en un array de salida
//! @param g_idata      Array de los valores de entrada
//! @param g_odata      Array de los valores de salida (valores reducidos)
///////////////////////////////////////////////////////////////////////////////////////////
__global__ void gpuFunc_Reduce(m_type* g_idata, m_type* g_odata){
    extern __shared__ m_type sdata[];
    // cada hilo carga un elemento desde memoria global hacia memoria shared
    unsigned int tid = threadIdx.x;
    unsigned int igl = blockIdx.x * blockDim.x + tid;
    sdata[tid] = g_idata[igl];
    __syncthreads();

    //Hacemos la reducción en memoria shared
    for (unsigned int s=blockDim.x/2; s>0; s/=2) {
        if (tid < s) {
 	  // Hacemos la reducción sumando los dos elementos que le tocan a este hilo
           sdata[tid] += sdata[tid+s];
        }
	__syncthreads();
    }
    
    // El hilo 0 de cada bloque escribe el resultado final de la reducción
    // en la memoria global del dispositivo pasada por parámetro (g_odata[])
    if (tid == 0)
        g_odata[blockIdx.x] = sdata[0];
}
#endif //_KERNELS_
